/*
    linenoise.qpp

    Qore Programming Language

    Copyright 2012 - 2022 Qore Technologies, s.r.o.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option any later version.

    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <errno.h>

extern "C" {
#include "linenoise.h"
}

/** @mainpage %Qore %Linenoise Module

    @tableofcontents

    @section jniintro jni Module Introduction

    The linenoise module provides readline-like functionality to Qore, allowing
    qore programs to manage comfortable user input in the command line.

    This module is released under a choice of two licenses:

    - LGPL 2.1
    - MIT (see COPYING.MIT in the source distribution for more information)

    The module is tagged as such in the module's header (meaning it can be
    loaded unconditionally regardless of how the Qore library was initialized).

    The underlying functionality is provided by Linenoise library, Steve Bennett's fork
    https://github.com/msteveb/linenoise
    of the original version
    https://github.com/antirez/linenoise

    @section basic_usage Basic Usage

    The core of functionality is the Qore::Linenoise::line() function. It waits for user input
    providing basic and advanced edit functionality.

    @code{.py}
%new-style
%require-types
%strict-args

%requires linenoise

while (True) {
    *string line = Linenoise::line("prompt> ");
    if (!exists line) {
        printf("^C signal caught. Exiting.\n");
        break;
    }

    # process the line value...
}
    @endcode

    @section history_handling History Handling

    Linenoise module keeps history in its internal structure. To add a history item
    use @ref Qore::Linenoise::history_add()

    Full history content can be obained with @ref Qore::Linenoise::history

    To handle history size use @ref Qore::Linenoise::history_set_max_len and
    @ref Qore::Linenoise::history_get_max_len

    History is not persistent accross processes (the memory is cleared when the process terminates).
    To store and load history content use @ref Qore::Linenoise::history_save()
    and @ref Qore::Linenoise::history_load()

    @section line_completion Line Completion

    Users start line completion by pressing the TAB key when in @ref Qore::Linenoise::line().

    An callback has to be used to provide line completion values.
    The callback is registered with @ref Qore::Linenoise::set_callback().

    @note The callback code must accept a \c string as an input argument and return a \c list.

    If the input argument isn't a string, a \c RUNTIME-OVERLOAD-ERROR exception will be thrown.

    If the return value isn't a list, the return will be ignored.

    Non string members of returned list will be converted to string (string representation
    of the list member will be used).

    <b>Example</b>
    @code{.py}
# h<TAB> is expanded to string: history
list sub my_callback(string value) {
    list ret = ();
    if (value == 'h')
        push ret, 'history';

    return ret;
}

Linenoise::set_callback(\my_callback());
    @endcode

    @section key_shortcuts Keyboard Shortcuts

    <table>
    <tr><th>Shortcut</th><th>Alternative</th><th>Description</th></tr>
    <tr><td>\c Enter</td><td>\c Return</td><td>submit the user input</td></tr>
    <tr><td>\c Tab</td><td>&nbsp;</td><td>starts a code completion if the callback is defined, @ref line_completion</td></tr>
    <tr><td>\c Up</td><td>\c Ctrl+P</td><td>history move, one step up</td></tr>
    <tr><td>\c Down</td><td>\c Ctrl+N</td><td>history move, one step down</td></tr>
    <tr><td>\c PgUp<td>&nbsp;</td><td>move to start of history</td></tr>
    <tr><td>\c PgDown<td>&nbsp;</td><td>move to the end of history = current text</td></tr>
    <tr><td>\c Left</td><td>\c Ctrl+B</td><td>move by one char left</td></tr>
    <tr><td>\c Right</td><td>\c Ctrl+F</td><td>move by one char right</td></tr>
    <tr><td>\c Home</td><td>\c Ctrl+A</td><td>go to the start of the line</td></tr>
    <tr><td>\c End</td><td>\c Ctrl+E</td><td>go to the end of the line</td></tr>
    <tr><td>\c Ctrl+C<td>&nbsp;</td><td>keyboard interrupt, @ref nothing "NOTHING" is returned from Qore::Linenoise::line() call</td></tr>
    <tr><td>\c Backspace</td><td>\c Ctrl+H</td><td>delete one char before</td></tr>
    <tr><td>\c Ctrl+D<td>&nbsp;</td><td>like a Ctrl+C when there is no user content in edit buffer, otherwise like a DELETE</td></tr>
    <tr><td>\c Ctrl+W<td>&nbsp;</td><td>delete word at left, save deleted chars</td></tr>
    <tr><td>\c Ctrl+R<td>&nbsp;</td><td>display the reverse-i-search prompt and process chars. See an additional controls for @ref find_key_shortcut</td></tr>
    <tr><td>\c Ctrl+T<td>&nbsp;</td><td>if cursor is at end, transpose the previous two chars</td></tr>
    <tr><td>\c Ctrl+V<td>&nbsp;</td><td>handle ^V sequence. Add ^V than waits for any other input or cancel ^V if it's in ^V mode already</td></tr>
    <tr><td>\c Ctrl+U<td>&nbsp;</td><td>delete to beginning of line, save deleted chars</td></tr>
    <tr><td>\c Ctrl+K<td>&nbsp;</td><td>delete from current to end of line, save deleted chars</td></tr>
    <tr><td>\c Ctrl+Y<td>&nbsp;</td><td>insert saved chars at current position</td></tr>
    <tr><td>\c Ctrl+L<td>&nbsp;</td><td>clear screen</td></tr>
    </table>

    @subsection find_key_shortcut Search Keyboard Shortcuts

    These shortcuts are valid in reverse-i-search (\c Ctrl+R) mode

    <table>
    <tr><th>Shortcut</th><th>Alternative</th><th>Description</th></tr>
    <tr><td>\c Delete</td><td>\c Ctrl+H</td><td>delete, return to normal operation mode if the line is empty</td></tr>
    <tr><td>\c Up</td><td>\c Ctrl+P</td><td>Search for the previous (earlier) match</td></tr>
    <tr><td>\c Down</td><td>\c Ctrl+N</td><td>Search for the next (later) match</td></tr>
    <tr><td>\c Ctrl+G</td><td>\c Ctrl+C</td><td>terminates the search with no effect</td></tr>
    <tr><td>\c Ctrl+J</td><td>&nbsp;</td><td>terminates the search leaving the buffer in place</td></tr>
    </table>

    @section jnireleasenotes jni Module Release Notes

    @subsection linenoise_1_0_1 linenoise Module Version 1.0.1
    - updated to build with \c qpp from %Qore 1.12.4+

    @subsection linenoise_1_0_0 linenoise Module Version 1.0.0
    - initial release
*/

static QoreNamespace LinenoiseNS("Linenoise");

void init_linenoise_functions(QoreNamespace& ns);

QoreThreadLock lock;
ResolvedCallReferenceNode* callbackNode = nullptr;

void qore_lineoise_completion_cb(const char *input, linenoiseCompletions *qter) {
//    AutoLocker al(lock);

    if (!callbackNode) {
        //printf("callback ignored, node is not set\n");
        return;
    }

    ExceptionSink xsink;

    ValueHolder ret(&xsink);
    {
       ReferenceHolder<QoreListNode> args(new QoreListNode(autoTypeInfo), &xsink);
       args->push(new QoreStringNode(input), nullptr);
       ret = callbackNode->execValue(*args, &xsink);
    }

    if (ret->getType() != NT_LIST) {
        printf("Callback implementation does not return list\n");
        return;
    }

    QoreListNode* lst = ret->get<QoreListNode>();
    for (size_t i = 0; i < lst->size(); ++i) {
        QoreStringValueHelper str(lst->retrieveEntry(i));
        linenoiseAddCompletion(qter, str->c_str());
    }
}

QoreStringNode * linenoise_module_init() {
    init_linenoise_functions(LinenoiseNS);

    linenoiseSetCompletionCallback(qore_lineoise_completion_cb);

    return 0;
}

void linenoise_module_ns_init(QoreNamespace *rns, QoreNamespace *qns) {
    qns->addNamespace(LinenoiseNS.copy());
}

void linenoise_module_delete() {
    if (callbackNode) {
        ExceptionSink xsink;
        callbackNode->deref(&xsink);
    }
}

// qore module symbols
DLLEXPORT char qore_module_name[] = "linenoise";
DLLEXPORT char qore_module_version[] = PACKAGE_VERSION;
DLLEXPORT char qore_module_description[] = "Linenoise module";
DLLEXPORT char qore_module_author[] = "Petr Vanek";
DLLEXPORT char qore_module_url[] = "http://qore.org";
DLLEXPORT int qore_module_api_major = QORE_MODULE_API_MAJOR;
DLLEXPORT int qore_module_api_minor = QORE_MODULE_API_MINOR;
DLLEXPORT qore_module_init_t qore_module_init = linenoise_module_init;
DLLEXPORT qore_module_ns_init_t qore_module_ns_init = linenoise_module_ns_init;
DLLEXPORT qore_module_delete_t qore_module_delete = linenoise_module_delete;
DLLEXPORT qore_license_t qore_module_license = QL_LGPL;
# 236 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
# 241 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
# 276 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
# 289 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
# 299 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
# 309 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
# 326 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
# 343 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
# 350 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
# 372 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
# 382 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
// int columns() {}
static QoreValue f_columns(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 376 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
    AutoLocker al(lock);

    return linenoiseColumns();
}

// list<auto> history() {}
static QoreValue f_history(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 358 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
    AutoLocker al(lock);

    int len;
    char** data = linenoiseHistory(&len);
    QoreListNode*ret = new QoreListNode(autoTypeInfo);

    for (int i = 0; i < len; ++i) {
        ret->push(new QoreStringNode(data[i]), xsink);
    }

    return ret;
}

// int history_add(string line) {}
static QoreValue f_history_add_Vs(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* line = HARD_QORE_VALUE_STRING(args, 0);
# 283 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
    AutoLocker al(lock);

    return linenoiseHistoryAdd(line->getBuffer());
}

// nothing history_free() {}
static QoreValue f_history_free(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 344 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
    AutoLocker al(lock);

    linenoiseHistoryFree();
    return QoreValue();
}

// int history_get_max_len() {}
static QoreValue f_history_get_max_len(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
# 303 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
    AutoLocker al(lock);

    return linenoiseHistoryGetMaxLen();
}

// nothing history_load(string filename) {}
static QoreValue f_history_load_Vs(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* filename = HARD_QORE_VALUE_STRING(args, 0);
# 334 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
    AutoLocker al(lock);

    if (linenoiseHistoryLoad(filename->getBuffer())) {
        xsink->raiseException("LINENOISE-HISTORY-LOAD-ERROR",
                              "error: %d; description: %s", errno, strerror(errno));
    }
    return QoreValue();
}

// nothing history_save(string filename) {}
static QoreValue f_history_save_Vs(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* filename = HARD_QORE_VALUE_STRING(args, 0);
# 317 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
    AutoLocker al(lock);

    if (linenoiseHistorySave(filename->getBuffer())) {
        xsink->raiseException("LINENOISE-HISTORY-SAVE-ERROR",
                              "error: %d; description: %s", errno, strerror(errno));
    }
    return QoreValue();
}

// int history_set_max_len(int len) {}
static QoreValue f_history_set_max_len_Vi(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    int64 len = HARD_QORE_VALUE_INT(args, 0);
# 293 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
    AutoLocker al(lock);

    return linenoiseHistorySetMaxLen(len);
}

// string line(string prompt) {}
static QoreValue f_line_Vs(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const QoreStringNode* prompt = HARD_QORE_VALUE_STRING(args, 0);
# 263 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
    AutoLocker al(lock);

    char * ret = linenoise(prompt->getBuffer());
    if (ret) {
        QoreStringNode *s = new QoreStringNode(ret);
        free(ret);
        return s;
    }
    else
        return nothing();
}

// nothing set_callback(code callback) {}
static QoreValue f_set_callback_Vc(const QoreListNode* args, q_rt_flags_t rtflags, ExceptionSink* xsink) {
    const ResolvedCallReferenceNode* callback = get_param_value(args, 0).get<const ResolvedCallReferenceNode>();
# 393 "/home/abuild/rpmbuild/BUILD/module-linenoise-1.0.1/src/linenoise.qpp"
    AutoLocker al(lock);

    if (callbackNode)
        callbackNode->deref(xsink);
    callbackNode = callback->refRefSelf();
    return QoreValue();
}


DLLLOCAL void init_linenoise_functions(QoreNamespace& ns) {
    // int columns() {}
    ns.addBuiltinVariant("columns", (q_func_n_t)f_columns, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);
    // list<auto> history() {}
    ns.addBuiltinVariant("history", (q_func_n_t)f_history, QCF_NO_FLAGS, QDOM_DEFAULT, autoListTypeInfo);
    // int history_add(string line) {}
    ns.addBuiltinVariant("history_add", (q_func_n_t)f_history_add_Vs, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "line");
    // nothing history_free() {}
    ns.addBuiltinVariant("history_free", (q_func_n_t)f_history_free, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);
    // int history_get_max_len() {}
    ns.addBuiltinVariant("history_get_max_len", (q_func_n_t)f_history_get_max_len, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);
    // nothing history_load(string filename) {}
    ns.addBuiltinVariant("history_load", (q_func_n_t)f_history_load_Vs, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "filename");
    // nothing history_save(string filename) {}
    ns.addBuiltinVariant("history_save", (q_func_n_t)f_history_save_Vs, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "filename");
    // int history_set_max_len(int len) {}
    ns.addBuiltinVariant("history_set_max_len", (q_func_n_t)f_history_set_max_len_Vi, QCF_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, bigIntTypeInfo, QORE_PARAM_NO_ARG, "len");
    // string line(string prompt) {}
    ns.addBuiltinVariant("line", (q_func_n_t)f_line_Vs, QCF_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG, "prompt");
    // nothing set_callback(code callback) {}
    ns.addBuiltinVariant("set_callback", (q_func_n_t)f_set_callback_Vc, QCF_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, codeTypeInfo, QORE_PARAM_NO_ARG, "callback");

}
